# -*- coding: utf-8 -*-

import click

from port_ocean.cli.commands.main import cli_start, print_logo, console
from port_ocean.config.settings import LogLevelType


@cli_start.command()
@click.argument("path", default=".", type=click.Path(exists=True))
@click.option(
    "-l",
    "--log-level",
    "log_level",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    default="DEBUG",
    help="""Set the logging level for the integration.
            Supported levels are DEBUG, INFO, WARNING, ERROR,
            and CRITICAL. If not specified, the default level
            is DEBUG.""",
)
@click.option(
    "-p",
    "--port",
    "port",
    type=int,
    default=8000,
    help="""Set the port for the integration to run on.
            If not specified, the default port is 8000.""",
)
@click.option(
    "-i",
    "--initialize-port-resources",
    "initialize_port_resources",
    type=bool,
    help="""Set to true to create default resources on installation.
            If not specified, the default value is false.""",
)
def sail(
    path: str,
    log_level: LogLevelType,
    port: int,
    initialize_port_resources: bool | None,
) -> None:
    """
    Runs the integration in the given PATH. if no PATH is provided, the current directory will be used.

    PATH: Path to the integration.
    """
    from port_ocean import run

    print_logo()

    console.print("Setting sail... ⛵️⚓️⛵️⚓️ All hands on deck! ⚓️")
    run(path, log_level, port, initialize_port_resources)



import asyncio
from itertools import chain

from loguru import logger

from port_ocean.clients.port.types import UserAgentType
from port_ocean.context.event import event
from port_ocean.core.handlers.entities_state_applier.base import (
    BaseEntitiesStateApplier,
)
from port_ocean.core.handlers.entities_state_applier.port.get_related_entities import (
    get_related_entities,
)
from port_ocean.core.handlers.entities_state_applier.port.order_by_entities_dependencies import (
    order_by_entities_dependencies,
)
from port_ocean.core.handlers.entities_state_applier.port.validate_entity_relations import (
    validate_entity_relations,
)
from port_ocean.core.handlers.entity_processor.base import EntityPortDiff
from port_ocean.core.models import Entity
from port_ocean.core.ocean_types import EntityDiff
from port_ocean.core.utils import is_same_entity, get_unique, get_port_diff
from port_ocean.exceptions.core import RelationValidationException


class HttpEntitiesStateApplier(BaseEntitiesStateApplier):
    async def _validate_delete_dependent_entities(self, entities: list[Entity]) -> None:
        logger.info("Validated deleted entities")
        if not event.port_app_config.delete_dependent_entities:
            dependent_entities = await asyncio.gather(
                *(
                    self.context.port_client.search_dependent_entities(entity)
                    for entity in entities
                )
            )
            new_dependent_entities = get_unique(
                [
                    entity
                    for entity in chain.from_iterable(dependent_entities)
                    if not any(is_same_entity(item, entity) for item in entities)
                ]
            )

            if new_dependent_entities:
                raise RelationValidationException(
                    f"Must enable delete_dependent_entities flag or delete all dependent entities: "
                    f" {[(dep.blueprint, dep.identifier) for dep in new_dependent_entities]}"
                )

    async def _validate_entity_diff(self, diff: EntityPortDiff) -> None:
        config = event.port_app_config
        await self._validate_delete_dependent_entities(diff.deleted)
        modified_or_created_entities = diff.modified + diff.created

        if modified_or_created_entities and not config.create_missing_related_entities:
            logger.info("Validating modified or created entities")

            await asyncio.gather(
                *(
                    self.context.port_client.validate_entity_payload(
                        entity,
                        {
                            "merge": config.enable_merge_entity,
                            "create_missing_related_entities": config.create_missing_related_entities,
                        },
                    )
                    for entity in modified_or_created_entities
                )
            )

        if not event.port_app_config.delete_dependent_entities:
            logger.info("Validating no relation blocks the operation")
            await validate_entity_relations(diff, self.context.port_client)

    async def _delete_diff(
        self,
        entities_to_delete: list[Entity],
        created_entities: list[Entity],
        user_agent_type: UserAgentType,
    ) -> None:
        if not entities_to_delete:
            return

        related_entities = await get_related_entities(
            created_entities, self.context.port_client
        )

        allowed_entities_to_delete = []

        for entity_to_delete in entities_to_delete:
            is_part_of_related = any(
                is_same_entity(entity, entity_to_delete) for entity in related_entities
            )
            is_part_of_created = any(
                is_same_entity(entity, entity_to_delete) for entity in created_entities
            )
            if is_part_of_related:
                if event.port_app_config.create_missing_related_entities:
                    logger.info(
                        f"Skipping entity {(entity_to_delete.identifier, entity_to_delete.blueprint)} because it is "
                        f"related to created entities and create_missing_related_entities is enabled"
                    )
                else:
                    allowed_entities_to_delete.append(entity_to_delete)
            elif not is_part_of_created:
                allowed_entities_to_delete.append(entity_to_delete)

        await self.delete(allowed_entities_to_delete, user_agent_type)

    async def apply_diff(
        self,
        entities: EntityDiff,
        user_agent_type: UserAgentType,
    ) -> None:
        diff = get_port_diff(entities["before"], entities["after"])

        logger.info(
            f"Updating entity diff (created: {len(diff.created)}, deleted: {len(diff.deleted)}, modified: {len(diff.modified)})"
        )
        await self._validate_entity_diff(diff)

        logger.info("Upserting new entities")
        await self.upsert(diff.created, user_agent_type)
        logger.info("Upserting modified entities")
        await self.upsert(diff.modified, user_agent_type)

        await self._delete_diff(
            diff.deleted, diff.created + diff.modified, user_agent_type
        )

    async def delete_diff(
        self,
        entities: EntityDiff,
        user_agent_type: UserAgentType,
    ) -> None:
        diff = get_port_diff(entities["before"], entities["after"])

        if not diff.deleted:
            return

        logger.info(
            f"Updating entity diff (created: {len(diff.created)}, deleted: {len(diff.deleted)}, modified: {len(diff.modified)})"
        )
        await self._validate_entity_diff(diff)

        await self._delete_diff(
            diff.deleted, diff.created + diff.modified, user_agent_type
        )

    async def upsert(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        logger.info(f"Upserting {len(entities)} entities")
        if event.port_app_config.create_missing_related_entities:
            await asyncio.gather(
                *(
                    self.context.port_client.upsert_entity(
                        entity,
                        event.port_app_config.get_port_request_options(),
                        user_agent_type,
                        should_raise=False,
                    )
                    for entity in entities
                ),
                return_exceptions=True,
            )
        else:
            ordered_created_entities = reversed(
                order_by_entities_dependencies(entities)
            )

            for entity in ordered_created_entities:
                await self.context.port_client.upsert_entity(
                    entity,
                    event.port_app_config.get_port_request_options(),
                    user_agent_type,
                    should_raise=False,
                )

    async def delete(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        logger.info(f"Deleting {len(entities)} entities")
        if event.port_app_config.delete_dependent_entities:
            await asyncio.gather(
                *(
                    self.context.port_client.delete_entity(
                        entity,
                        event.port_app_config.get_port_request_options(),
                        user_agent_type,
                        should_raise=False,
                    )
                    for entity in entities
                )
            )
        else:
            ordered_deleted_entities = order_by_entities_dependencies(entities)

            for entity in ordered_deleted_entities:
                await self.context.port_client.delete_entity(
                    entity,
                    event.port_app_config.get_port_request_options(),
                    user_agent_type,
                    should_raise=False,
                )



from typing import Literal, Any

from loguru import logger

from port_ocean.context.ocean import ocean
from port_ocean.core.event_listener.base import (
    BaseEventListener,
    EventListenerEvents,
    EventListenerSettings,
)
from port_ocean.core.event_listener.polling.utils import repeat_every


class PollingEventListenerSettings(EventListenerSettings):
    type: Literal["POLLING"]
    resync_on_start: bool = True
    interval: int = 60

    def to_request(self) -> dict[str, Any]:
        return {}


class PollingEventListener(BaseEventListener):
    def __init__(
        self,
        events: EventListenerEvents,
        event_listener_config: PollingEventListenerSettings,
    ):
        super().__init__(events)
        self.event_listener_config = event_listener_config
        self._last_updated_at = None

    async def start(self) -> None:
        logger.info(
            f"Setting up Polling event listener with interval: {self.event_listener_config.interval}"
        )

        @repeat_every(seconds=self.event_listener_config.interval)
        async def resync() -> None:
            logger.info(
                f"Polling event listener iteration after {self.event_listener_config.interval}. Checking for changes"
            )
            integration = await ocean.app.port_client.get_current_integration()
            last_updated_at = integration["updatedAt"]

            should_resync = (
                self._last_updated_at is not None
                or self.event_listener_config.resync_on_start
            ) and self._last_updated_at != last_updated_at

            if should_resync:
                logger.info("Detected change in integration, resyncing")
                self._last_updated_at = last_updated_at
                await self.events["on_resync"]({})

        # Execute resync repeatedly task
        await resync()



from graphlib import TopologicalSorter
from typing import Set

from port_ocean.core.models import Entity

Node = tuple[str, str]


def node(entity: Entity) -> Node:
    return entity.identifier, entity.blueprint


def order_by_entities_dependencies(entities: list[Entity]) -> list[Entity]:
    nodes: dict[Node, Set[Node]] = {}
    entities_map = {}

    for entity in entities:
        nodes[node(entity)] = set()
        entities_map[node(entity)] = entity

    for entity in entities:
        relation_target_ids: list[str] = sum(
            [
                identifiers if isinstance(identifiers, list) else [identifiers]
                for identifiers in entity.relations.values()
                if identifiers is not None
            ],
            [],
        )
        related_entities = [
            related for related in entities if related.identifier in relation_target_ids
        ]

        for related_entity in related_entities:
            nodes[node(entity)].add(node(related_entity))

    sort_op = TopologicalSorter(nodes)
    return [entities_map[item] for item in sort_op.static_order()]



from typing import Callable, Awaitable

from fastapi import Request, Response
from loguru import logger

from port_ocean.exceptions.api import BaseAPIException, InternalServerException
from .context.event import event_context, EventType
from .context.ocean import ocean
from .utils import get_time, generate_uuid


async def _handle_silently(
    call_next: Callable[[Request], Awaitable[Response]], request: Request
) -> Response:
    response: Response
    try:
        if request.url.path.startswith("/integration"):
            async with event_context(EventType.HTTP_REQUEST, trigger_type="request"):
                await ocean.integration.port_app_config_handler.get_port_app_config()
                response = await call_next(request)
        else:
            response = await call_next(request)

    except BaseAPIException as ex:
        response = ex.response()
        if response.status_code < 500:
            logger.bind(exception=str(ex)).info(
                "Request did not succeed due to client-side error"
            )
        else:
            logger.opt(exception=True).warning(
                "Request did not succeed due to server-side error"
            )

    except Exception:
        logger.opt(exception=True).error("Request failed due to unexpected error")
        response = InternalServerException().response()

    return response


async def request_handler(
    request: Request, call_next: Callable[[Request], Awaitable[Response]]
) -> Response:
    """Middleware used by FastAPI to process each request, featuring:

    - Contextualize request logs with a unique Request ID (UUID4) for each unique request.
    - Catch exceptions during the request handling. Translate custom API exceptions into responses,
      or treat (and log) unexpected exceptions.
    """
    start_time = get_time(seconds_precision=False)
    request_id = generate_uuid()

    with logger.contextualize(request_id=request_id):
        logger.bind(url=str(request.url), method=request.method).info("Request started")
        response = await _handle_silently(call_next, request)

        end_time = get_time(seconds_precision=False)
        time_elapsed = round(end_time - start_time, 5)
        response.headers["X-Request-ID"] = request_id
        response.headers["X-Process-Time"] = str(time_elapsed)
        logger.bind(
            time_elapsed=time_elapsed, response_status=response.status_code
        ).info("Request ended")
        return response



import httpx
from loguru import logger
from werkzeug.local import LocalStack, LocalProxy

_http_client: LocalStack[httpx.AsyncClient] = LocalStack()


def _get_http_client_context() -> httpx.AsyncClient:
    client = _http_client.top
    if client is None:
        client = httpx.AsyncClient()
        _http_client.push(client)

    return client


async_client: httpx.AsyncClient = LocalProxy(lambda: _get_http_client_context())  # type: ignore


def handle_status_code(
    response: httpx.Response, should_raise: bool = True, should_log: bool = True
) -> None:
    if should_log and response.is_error:
        logger.error(
            f"Request failed with status code: {response.status_code}, Error: {response.text}"
        )
    if should_raise:
        response.raise_for_status()



import json
from typing import Type, Any, Optional

from humps import decamelize
from pydantic import BaseModel, AnyUrl, create_model, Extra, parse_obj_as, validator
from pydantic.fields import ModelField


class Configuration(BaseModel, extra=Extra.allow):
    name: str
    type: str
    required: bool = False
    default: Optional[Any]


def dynamic_parse(value: Any, field: ModelField) -> Any:
    should_json_load = issubclass(field.annotation, dict) or issubclass(
        field.annotation, list
    )
    if isinstance(value, str) and should_json_load:
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            pass
    return value


def default_config_factory(configurations: Any) -> Type[BaseModel]:
    configurations = parse_obj_as(list[Configuration], configurations)
    fields: dict[str, tuple[Any, Any]] = {}

    for config in configurations:
        field_type: Type[Any]

        match config.type:
            case "object":
                field_type = dict
            case "url":
                field_type = AnyUrl
            case "string":
                field_type = str
            case "integer":
                field_type = int
            case "boolean":
                field_type = bool
            case _:
                raise ValueError(f"Unknown type: {config.type}")

        default = ... if config.required else None
        if config.default is not None:
            default = parse_obj_as(field_type, config.default)
        fields[decamelize(config.name)] = (
            field_type,
            default,
        )

    dynamic_model = create_model(  # type: ignore
        __model_name="Config",
        **fields,
        __validators__={"dynamic_parse": validator("*", pre=True)(dynamic_parse)},
    )
    return dynamic_model



import asyncio

from loguru import logger

from port_ocean.context.event import (
    event_context,
    EventType,
)
from port_ocean.context.ocean import PortOceanContext
from port_ocean.core.event_listener.factory import (
    EventListenerFactory,
)
from port_ocean.core.integrations.mixins.sync import SyncRawMixin, SyncMixin
from port_ocean.exceptions.core import IntegrationAlreadyStartedException


class BaseIntegration(SyncRawMixin, SyncMixin):
    def __init__(self, context: PortOceanContext):
        SyncRawMixin.__init__(self)
        SyncMixin.__init__(self)
        self.started = False
        self.context = context
        self.event_listener_factory = EventListenerFactory(
            context,
            self.context.config.integration.identifier,
            {"on_resync": self.sync_raw_all},
        )

    async def start(self) -> None:
        logger.info("Starting integration")
        if self.started:
            raise IntegrationAlreadyStartedException("Integration already started")

        if (
            not self.event_strategy["resync"]
            and self.__class__._on_resync == BaseIntegration._on_resync
        ):
            raise NotImplementedError("on_resync is not implemented")

        await self.initialize_handlers()

        logger.info("Initializing integration at port")
        await self.context.port_client.initialize_integration(
            self.context.config.integration.type,
            self.context.config.event_listener.to_request(),
        )

        self.started = True

        async with event_context(EventType.START, trigger_type="machine"):
            await asyncio.gather(
                *(listener() for listener in self.event_strategy["start"])
            )

        logger.info("Initializing event listener")
        event_listener = await self.event_listener_factory.create_event_listener()
        await event_listener.start()



from typing import Iterable, Any, TypeVar

from pydantic import parse_obj_as, ValidationError

from port_ocean.core.handlers.entity_processor.base import EntityPortDiff
from port_ocean.core.models import Entity
from port_ocean.core.ocean_types import RAW_RESULT
from port_ocean.exceptions.core import RawObjectValidationException

T = TypeVar("T", bound=tuple[list[Any], ...])


def zip_and_sum(collection: Iterable[T]) -> T:
    return tuple(sum(items, []) for items in zip(*collection))  # type: ignore


def validate_result(result: Any) -> RAW_RESULT:
    try:
        return parse_obj_as(list[dict[str, Any]], result)
    except ValidationError as e:
        raise RawObjectValidationException(f"Expected list[dict[str, Any]], Error: {e}")


def is_same_entity(first_entity: Entity, second_entity: Entity) -> bool:
    return (
        first_entity.identifier == second_entity.identifier
        and first_entity.blueprint == second_entity.blueprint
    )


def get_unique(array: list[Entity]) -> list[Entity]:
    result: list[Entity] = []
    for item in array:
        if all(not is_same_entity(item, seen_item) for seen_item in result):
            result.append(item)
    return result


def get_port_diff(
    before: Iterable[Entity],
    after: Iterable[Entity],
) -> EntityPortDiff:
    return EntityPortDiff(
        deleted=get_unique(
            [
                item
                for item in before
                if not any(is_same_entity(item, item_after) for item_after in after)
            ],
        ),
        created=get_unique(
            [
                item
                for item in after
                if not any(is_same_entity(item, item_before) for item_before in before)
            ],
        ),
        modified=get_unique(
            [
                item
                for item in after
                if any(is_same_entity(item, item_before) for item_before in before)
            ],
        ),
    )



import inspect
from pathlib import Path
from time import time
from typing import Callable, Any
from uuid import uuid4

import yaml


def get_time(seconds_precision: bool = True) -> float:
    """Return current time as Unix/Epoch timestamp, in seconds.
    :param seconds_precision: if True, return with seconds precision as integer (default).
                              If False, return with milliseconds precision as floating point number of seconds.
    """
    return time() if not seconds_precision else int(time())


def generate_uuid() -> str:
    """Return a UUID4 as string"""
    return str(uuid4())


def get_function_location(func: Callable[..., Any]) -> str:
    file_path = inspect.getsourcefile(func)
    line_number = inspect.getsourcelines(func)[1]
    return f"{file_path}:{line_number}"


def get_spec_file(path: Path = Path(".")) -> dict[str, Any] | None:
    try:
        return yaml.safe_load((path / ".port/spec.yaml").read_text())
    except FileNotFoundError:
        return None



# -*- coding: utf-8 -*-

import click

from port_ocean import __version__
from port_ocean.cli.commands.main import cli_start, console


@cli_start.command()
@click.option(
    "-s",
    "--short",
    "short",
    default=False,
    is_flag=True,
    required=False,
    help="Display only the short version number.",
)
def version(short: bool) -> None:
    """
    Displays the version of the Ocean package.
    """
    if short:
        console.print(__version__)
    else:
        console.print(f"🌊 Ocean version: {__version__}")



from loguru import logger

from port_ocean.context.ocean import PortOceanContext
from port_ocean.core.base import BaseWithContext
from port_ocean.core.event_listener import (
    HttpEventListener,
    KafkaEventListener,
    PollingEventListener,
)
from port_ocean.core.event_listener import (
    HttpEventListenerSettings,
    KafkaEventListenerSettings,
    PollingEventListenerSettings,
)
from port_ocean.core.event_listener.base import (
    BaseEventListener,
    EventListenerEvents,
)
from port_ocean.exceptions.core import UnsupportedEventListenerTypeException


class EventListenerFactory(BaseWithContext):
    def __init__(
        self,
        context: PortOceanContext,
        installation_id: str,
        events: EventListenerEvents,
    ):
        super().__init__(context)
        self.installation_id = installation_id
        self.events = events

    async def create_event_listener(self) -> BaseEventListener:
        wrapped_events: EventListenerEvents = {"on_resync": self.events["on_resync"]}
        event_listener: BaseEventListener
        config = self.context.config.event_listener
        _type = config.type.lower()
        assert_message = "Invalid event listener config, expected KafkaEventListenerSettings and got {0}"
        logger.info(f"Found event listener type: {_type}")

        match _type:
            case "kafka":
                assert isinstance(
                    config, KafkaEventListenerSettings
                ), assert_message.format(type(config))
                org_id = await self.context.port_client.get_org_id()
                event_listener = KafkaEventListener(
                    wrapped_events,
                    config,
                    org_id,
                    self.context.config.integration.identifier,
                    self.context.config.integration.type,
                )

            case "webhook":
                assert isinstance(
                    config, HttpEventListenerSettings
                ), assert_message.format(type(config))
                event_listener = HttpEventListener(wrapped_events, config)

            case "polling":
                assert isinstance(
                    config, PollingEventListenerSettings
                ), assert_message.format(type(config))
                event_listener = PollingEventListener(wrapped_events, config)

            case _:
                raise UnsupportedEventListenerTypeException(
                    f"Event listener {_type} not supported"
                )

        return event_listener



from typing import Any

import httpx
from loguru import logger

from port_ocean.clients.port.authentication import PortAuthentication
from port_ocean.clients.utils import handle_status_code
from port_ocean.core.models import Blueprint


class BlueprintClientMixin:
    def __init__(self, auth: PortAuthentication, client: httpx.AsyncClient):
        self.auth = auth
        self.client = client

    async def get_blueprint(self, identifier: str) -> Blueprint:
        logger.info(f"Fetching blueprint with id: {identifier}")
        response = await self.client.get(
            f"{self.auth.api_url}/blueprints/{identifier}",
            headers=await self.auth.headers(),
        )
        handle_status_code(response)
        return Blueprint.parse_obj(response.json()["blueprint"])

    async def create_blueprint(self, raw_blueprint: dict[str, Any]) -> None:
        logger.info(f"Creating blueprint with id: {raw_blueprint.get('identifier')}")
        headers = await self.auth.headers()
        response = await self.client.post(
            f"{self.auth.api_url}/blueprints", headers=headers, json=raw_blueprint
        )
        handle_status_code(response)
        if response.is_success:
            return response.json()["blueprint"]

    async def patch_blueprint(
        self, identifier: str, raw_blueprint: dict[str, Any]
    ) -> None:
        logger.info(f"Patching blueprint with id: {identifier}")
        headers = await self.auth.headers()
        response = await self.client.patch(
            f"{self.auth.api_url}/blueprints/{identifier}",
            headers=headers,
            json=raw_blueprint,
        )
        handle_status_code(response)

    async def delete_blueprint(
        self, identifier: str, should_raise: bool = False
    ) -> None:
        logger.info(f"Deleting blueprint with id: {identifier}")
        headers = await self.auth.headers()
        response = await self.client.delete(
            f"{self.auth.api_url}/blueprints/{identifier}",
            headers=headers,
        )
        handle_status_code(response, should_raise)

    async def create_action(
        self, blueprint_identifier: str, action: dict[str, Any]
    ) -> None:
        logger.info(f"Creating action: {action}")
        response = await self.client.post(
            f"{self.auth.api_url}/blueprints/{blueprint_identifier}/actions",
            json=action,
            headers=await self.auth.headers(),
        )

        handle_status_code(response)

    async def create_scorecard(
        self,
        blueprint_identifier: str,
        scorecard: dict[str, Any],
    ) -> None:
        logger.info(f"Creating scorecard: {scorecard}")
        response = await self.client.post(
            f"{self.auth.api_url}/blueprints/{blueprint_identifier}/scorecards",
            json=scorecard,
            headers=await self.auth.headers(),
        )

        handle_status_code(response)



from typing import Any, Literal

from pydantic import BaseSettings, BaseModel, Extra, AnyHttpUrl, parse_obj_as

from port_ocean.config.base import BaseOceanSettings
from port_ocean.core.event_listener import EventListenerSettingsType

LogLevelType = Literal["ERROR", "WARNING", "INFO", "DEBUG", "CRITICAL"]


class ApplicationSettings(BaseSettings):
    log_level: LogLevelType = "INFO"
    port: int = 8000

    class Config:
        env_prefix = "APPLICATION__"
        env_file = ".env"
        env_file_encoding = "utf-8"

        @classmethod
        def customise_sources(cls, init_settings, env_settings, *_, **__):  # type: ignore
            return env_settings, init_settings


class PortSettings(BaseModel, extra=Extra.allow):
    client_id: str
    client_secret: str
    base_url: AnyHttpUrl = parse_obj_as(AnyHttpUrl, "https://api.getport.io")


class IntegrationSettings(BaseModel, extra=Extra.allow):
    identifier: str
    type: str
    config: dict[str, Any]


class IntegrationConfiguration(BaseOceanSettings, extra=Extra.allow):
    port: PortSettings
    event_listener: EventListenerSettingsType
    batch_work_size: int = 20
    initialize_port_resources: bool = False
    integration: IntegrationSettings



from abc import abstractmethod

from port_ocean.clients.port.types import UserAgentType
from port_ocean.core.base import BaseWithContext
from port_ocean.core.models import Entity
from port_ocean.core.ocean_types import EntityDiff


class BaseEntitiesStateApplier(BaseWithContext):
    @abstractmethod
    async def apply_diff(
        self,
        entities: EntityDiff,
        user_agent: UserAgentType,
    ) -> None:
        pass

    @abstractmethod
    async def delete_diff(
        self,
        entities: EntityDiff,
        user_agent: UserAgentType,
    ) -> None:
        pass

    @abstractmethod
    async def upsert(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        pass

    @abstractmethod
    async def delete(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        pass



import asyncio
import inspect
import typing
from typing import Any, Awaitable, Callable

from loguru import logger

from port_ocean.clients.port.types import UserAgentType
from port_ocean.context.event import TriggerType, event_context, EventType
from port_ocean.context.ocean import ocean
from port_ocean.core.handlers.port_app_config.models import ResourceConfig
from port_ocean.core.integrations.mixins.events import EventsMixin
from port_ocean.core.integrations.mixins.handler import HandlerMixin
from port_ocean.core.integrations.mixins.utils import (
    resync_function_wrapper,
    resync_generator_wrapper,
)
from port_ocean.core.models import Entity
from port_ocean.core.ocean_types import (
    RawEntityDiff,
    EntityDiff,
    RESYNC_RESULT,
    RAW_RESULT,
    RESYNC_EVENT_LISTENER,
    ASYNC_GENERATOR_RESYNC_TYPE,
)
from port_ocean.core.utils import zip_and_sum
from port_ocean.exceptions.core import OceanAbortException


class SyncMixin(HandlerMixin):
    def __init__(self) -> None:
        HandlerMixin.__init__(self)

    async def update_diff(
        self,
        desired_state: EntityDiff,
        user_agent_type: UserAgentType,
    ) -> None:
        await self.entities_state_applier.apply_diff(
            {"before": desired_state["before"], "after": desired_state["after"]},
            user_agent_type,
        )

    async def register(
        self,
        entities: list[Entity],
        user_agent_type: UserAgentType,
    ) -> None:
        await self.entities_state_applier.upsert(entities, user_agent_type)
        logger.info("Finished registering change")

    async def unregister(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        await self.entities_state_applier.delete(entities, user_agent_type)
        logger.info("Finished unregistering change")

    async def sync(
        self,
        entities: list[Entity],
        user_agent_type: UserAgentType,
    ) -> None:
        entities_at_port = await ocean.port_client.search_entities(user_agent_type)

        await self.entities_state_applier.upsert(entities, user_agent_type)
        await self.entities_state_applier.delete_diff(
            {"before": entities_at_port, "after": entities}, user_agent_type
        )

        logger.info("Finished syncing change")


class SyncRawMixin(HandlerMixin, EventsMixin):
    def __init__(self) -> None:
        HandlerMixin.__init__(self)
        EventsMixin.__init__(self)

    async def _on_resync(self, kind: str) -> RAW_RESULT:
        raise NotImplementedError("on_resync must be implemented")

    async def _calculate_raw(
        self, raw_diff: list[tuple[ResourceConfig, RawEntityDiff]]
    ) -> list[EntityDiff]:
        logger.info("Calculating diff in entities between states")
        return await asyncio.gather(
            *(
                self.entity_processor.parse_items(mapping, results)
                for mapping, results in raw_diff
            )
        )

    async def _get_resource_raw_results(
        self, resource_config: ResourceConfig
    ) -> tuple[RESYNC_RESULT, list[Exception]]:
        logger.info(f"Fetching {resource_config.kind} resync results")
        tasks: list[Awaitable[RAW_RESULT]] = []
        with logger.contextualize(kind=resource_config.kind):
            fns: list[RESYNC_EVENT_LISTENER] = [
                *self.event_strategy["resync"][resource_config.kind],
                *self.event_strategy["resync"][None],
            ]

            if self.__class__._on_resync != SyncRawMixin._on_resync:
                fns.append(self._on_resync)

            results: RESYNC_RESULT = []
            for task in fns:
                if inspect.isasyncgenfunction(task):
                    results.append(resync_generator_wrapper(task, resource_config.kind))
                else:
                    task = typing.cast(Callable[[str], Awaitable[RAW_RESULT]], task)
                    tasks.append(resync_function_wrapper(task, resource_config.kind))

            logger.info(
                f"Found {len(tasks) + len(results)} resync tasks for {resource_config.kind}"
            )

            results_with_error: list[RAW_RESULT | Exception] = await asyncio.gather(
                *tasks,
                return_exceptions=True,
            )
            results.extend(
                sum(
                    [
                        result
                        for result in results_with_error
                        if not isinstance(result, Exception)
                    ],
                    [],
                )
            )

            errors = [
                result for result in results_with_error if isinstance(result, Exception)
            ]

            logger.info(
                f"Triggered {len(tasks)} tasks for {resource_config.kind}, failed: {len(errors)}"
            )
            return results, errors

    async def _register_resource_raw(
        self,
        resource: ResourceConfig,
        results: list[dict[Any, Any]],
        user_agent_type: UserAgentType,
    ) -> list[Entity]:
        objects_diff = await self._calculate_raw(
            [
                (
                    resource,
                    {
                        "before": [],
                        "after": results,
                    },
                )
            ]
        )

        entities_after: list[Entity] = objects_diff[0]["after"]
        await self.entities_state_applier.upsert(entities_after, user_agent_type)
        return entities_after

    async def _unregister_resource_raw(
        self,
        resource: ResourceConfig,
        results: list[dict[Any, Any]],
        user_agent_type: UserAgentType,
    ) -> list[Entity]:
        objects_diff = await self._calculate_raw(
            [
                (
                    resource,
                    {
                        "before": results,
                        "after": [],
                    },
                )
            ]
        )

        entities_after: list[Entity] = objects_diff[0]["before"]
        await self.entities_state_applier.delete(entities_after, user_agent_type)
        logger.info("Finished unregistering change")
        return entities_after

    async def _register_in_batches(
        self,
        resource_config: ResourceConfig,
        user_agent_type: UserAgentType,
        batch_work_size: int,
    ) -> tuple[list[Entity], list[Exception]]:
        results, errors = await self._get_resource_raw_results(resource_config)
        async_generators: list[ASYNC_GENERATOR_RESYNC_TYPE] = []
        raw_results: RAW_RESULT = []
        for result in results:
            if isinstance(result, dict):
                raw_results.append(result)
            else:
                async_generators.append(result)

        batches = [
            raw_results[i : i + batch_work_size]
            for i in range(0, len(raw_results), batch_work_size)
        ]

        registered_entities_results = await asyncio.gather(
            *(
                self._register_resource_raw(resource_config, batch, user_agent_type)
                for batch in batches
            )
        )
        for generator in async_generators:
            try:
                async for items in generator:
                    registered_entities_results.append(
                        await self._register_resource_raw(
                            resource_config, items, user_agent_type
                        )
                    )
            except* OceanAbortException as error:
                errors.append(error)

        entities: list[Entity] = sum(registered_entities_results, [])
        logger.info(
            f"Finished registering change for {len(results)} raw results for kind: {resource_config.kind}. {len(entities)} entities were affected"
        )
        return entities, errors

    async def register_raw(
        self,
        kind: str,
        results: list[dict[Any, Any]],
        user_agent_type: UserAgentType,
    ) -> list[Entity]:
        logger.info(f"Registering state for {kind}")
        config = await self.port_app_config_handler.get_port_app_config()
        resource_mappings = [
            resource for resource in config.resources if resource.kind == kind
        ]

        return await asyncio.gather(
            *(
                self._register_resource_raw(resource, results, user_agent_type)
                for resource in resource_mappings
            )
        )

    async def unregister_raw(
        self,
        kind: str,
        results: list[dict[Any, Any]],
        user_agent_type: UserAgentType,
    ) -> list[Entity]:
        logger.info(f"Registering state for {kind}")
        config = await self.port_app_config_handler.get_port_app_config()
        resource_mappings = [
            resource for resource in config.resources if resource.kind == kind
        ]

        return await asyncio.gather(
            *(
                self._unregister_resource_raw(resource, results, user_agent_type)
                for resource in resource_mappings
            )
        )

    async def update_raw_diff(
        self,
        kind: str,
        raw_desired_state: RawEntityDiff,
        user_agent_type: UserAgentType,
    ) -> None:
        logger.info(f"Updating state for {kind}")
        config = await self.port_app_config_handler.get_port_app_config()
        resource_mappings = [
            resource for resource in config.resources if resource.kind == kind
        ]

        with logger.contextualize(kind=kind):
            logger.info(f"Found {len(resource_mappings)} resources for {kind}")

            objects_diff = await self._calculate_raw(
                [(mapping, raw_desired_state) for mapping in resource_mappings]
            )

            entities_before, entities_after = zip_and_sum(
                (
                    (entities_change["before"], entities_change["after"])
                    for entities_change in objects_diff
                )
            )

            await self.entities_state_applier.apply_diff(
                {"before": entities_before, "after": entities_after}, user_agent_type
            )

    async def sync_raw_all(
        self,
        _: dict[Any, Any] | None = None,
        trigger_type: TriggerType = "machine",
        user_agent_type: UserAgentType = UserAgentType.exporter,
        silent: bool = True,
    ) -> None:
        logger.info("Resync was triggered")

        async with event_context(EventType.RESYNC, trigger_type=trigger_type):
            app_config = await self.port_app_config_handler.get_port_app_config()

            entities_at_port = await ocean.port_client.search_entities(user_agent_type)

            creation_results: list[tuple[list[Entity], list[Exception]]] = []
            for resource in app_config.resources:
                # create event context per resource kind, so resync method could have access to the resource config
                # as we might have multiple resources with the same kind name
                async with event_context(
                    EventType.RESYNC,
                    trigger_type=trigger_type,
                    resource_config=resource,
                    attributes={"resource_kind": resource.kind},
                ):
                    creation_results.append(
                        await self._register_in_batches(
                            resource, user_agent_type, ocean.config.batch_work_size
                        )
                    )
            flat_created_entities, errors = zip_and_sum(creation_results) or [[], []]

            if errors:
                message = f"Resync failed with {len(errors)}. Skipping delete phase due to incomplete state"
                error_group = ExceptionGroup(
                    f"Resync failed with {len(errors)}. Skipping delete phase due to incomplete state",
                    errors,
                )
                if not silent:
                    raise error_group

                logger.error(message, exc_info=error_group)
            else:
                await self.entities_state_applier.delete_diff(
                    {"before": entities_at_port, "after": flat_created_entities},
                    user_agent_type,
                )



from dataclasses import dataclass
from typing import Callable, TYPE_CHECKING, Any

from fastapi import APIRouter
from werkzeug.local import LocalProxy, LocalStack

from port_ocean.clients.port.types import UserAgentType
from port_ocean.core.models import Entity
from port_ocean.core.ocean_types import (
    RESYNC_EVENT_LISTENER,
    START_EVENT_LISTENER,
    RawEntityDiff,
    EntityDiff,
)
from port_ocean.exceptions.context import PortOceanContextNotFoundError

if TYPE_CHECKING:
    from port_ocean.config.settings import IntegrationConfiguration
    from port_ocean.core.integrations.base import BaseIntegration
    from port_ocean.ocean import Ocean
    from port_ocean.clients.port.client import PortClient


@dataclass
class PortOceanContext:
    app: "Ocean"

    @property
    def config(self) -> "IntegrationConfiguration":
        return self.app.config

    @property
    def router(self) -> APIRouter:
        return self.app.integration_router

    @property
    def integration(self) -> "BaseIntegration":
        return self.app.integration

    @property
    def integration_config(self) -> dict[str, Any]:
        return self.app.config.integration.config

    @property
    def port_client(self) -> "PortClient":
        return self.app.port_client

    def on_resync(
        self,
        kind: str | None = None,
    ) -> Callable[[RESYNC_EVENT_LISTENER], RESYNC_EVENT_LISTENER]:
        def wrapper(function: RESYNC_EVENT_LISTENER) -> RESYNC_EVENT_LISTENER:
            return self.integration.on_resync(function, kind)

        return wrapper

    def on_start(self) -> Callable[[START_EVENT_LISTENER], START_EVENT_LISTENER]:
        def wrapper(function: START_EVENT_LISTENER) -> START_EVENT_LISTENER:
            return self.integration.on_start(function)

        return wrapper

    async def update_raw_diff(
        self,
        kind: str,
        raw_diff: RawEntityDiff,
        user_agent_type: UserAgentType = UserAgentType.exporter,
    ) -> None:
        await self.integration.update_raw_diff(kind, raw_diff, user_agent_type)

    async def update_diff(
        self,
        diff: EntityDiff,
        user_agent_type: UserAgentType = UserAgentType.exporter,
    ) -> None:
        await self.integration.update_diff(diff, user_agent_type)

    async def register_raw(
        self,
        kind: str,
        change: list[dict[str, Any]],
        user_agent_type: UserAgentType = UserAgentType.exporter,
    ) -> None:
        await self.integration.register_raw(kind, change, user_agent_type)

    async def unregister_raw(
        self,
        kind: str,
        change: list[dict[str, Any]],
        user_agent_type: UserAgentType = UserAgentType.exporter,
    ) -> None:
        await self.integration.unregister_raw(kind, change, user_agent_type)

    async def register(
        self,
        entities: list[Entity],
        user_agent_type: UserAgentType = UserAgentType.exporter,
    ) -> None:
        await self.integration.register(entities, user_agent_type)

    async def unregister(
        self,
        entities: list[Entity],
        user_agent_type: UserAgentType = UserAgentType.exporter,
    ) -> None:
        await self.integration.unregister(entities, user_agent_type)

    async def sync(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        await self.integration.sync(entities, user_agent_type)

    async def sync_raw_all(self) -> None:
        await self.integration.sync_raw_all(trigger_type="manual")


_port_ocean_context_stack: LocalStack[PortOceanContext] = LocalStack()


def initialize_port_ocean_context(ocean_app: "Ocean") -> None:
    """
    This Function initializes the PortOcean context and pushes it into the LocalStack().
    """
    _port_ocean_context_stack.push(PortOceanContext(app=ocean_app))


def _get_port_ocean_context() -> PortOceanContext:
    """
    Get the PortOcean context from the current thread.
    """
    port_ocean_context = _port_ocean_context_stack.top
    if port_ocean_context is None:
        raise PortOceanContextNotFoundError(
            "You must first initialize PortOcean in order to use it"
        )

    return port_ocean_context


ocean: PortOceanContext = LocalProxy(lambda: _get_port_ocean_context())  # type: ignore



import warnings
from importlib.metadata import version

warnings.filterwarnings("ignore", category=FutureWarning)

from .ocean import Ocean  # noqa: E402
from .run import run  # noqa: E402

__version__ = version("port-ocean")


__all__ = ["Ocean", "run", "__version__"]



from typing import Any, TYPE_CHECKING, Optional

import httpx
from loguru import logger
from starlette import status

from port_ocean.clients.port.authentication import PortAuthentication
from port_ocean.clients.utils import handle_status_code

if TYPE_CHECKING:
    from port_ocean.core.handlers.port_app_config.models import PortAppConfig


class IntegrationClientMixin:
    def __init__(
        self,
        integration_identifier: str,
        auth: PortAuthentication,
        client: httpx.AsyncClient,
    ):
        self.integration_identifier = integration_identifier
        self.auth = auth
        self.client = client

    async def _get_current_integration(self) -> httpx.Response:
        logger.info(f"Fetching integration with id: {self.integration_identifier}")
        response = await self.client.get(
            f"{self.auth.api_url}/integration/{self.integration_identifier}",
            headers=await self.auth.headers(),
        )
        return response

    async def get_current_integration(
        self, should_raise: bool = True, should_log: bool = True
    ) -> dict[str, Any]:
        response = await self._get_current_integration()
        handle_status_code(response, should_raise, should_log)
        return response.json()["integration"]

    async def create_integration(
        self,
        _type: str,
        changelog_destination: dict[str, Any],
        port_app_config: Optional["PortAppConfig"] = None,
    ) -> None:
        logger.info(f"Creating integration with id: {self.integration_identifier}")
        headers = await self.auth.headers()
        json = {
            "installationId": self.integration_identifier,
            "installationAppType": _type,
            "changelogDestination": changelog_destination,
            "config": {},
        }
        if port_app_config:
            json["config"] = port_app_config.to_request()
        response = await self.client.post(
            f"{self.auth.api_url}/integration", headers=headers, json=json
        )
        handle_status_code(response)

    async def patch_integration(
        self,
        _type: str | None = None,
        changelog_destination: dict[str, Any] | None = None,
        port_app_config: Optional["PortAppConfig"] = None,
    ) -> None:
        logger.info(f"Updating integration with id: {self.integration_identifier}")
        headers = await self.auth.headers()
        json: dict[str, Any] = {}
        if _type:
            json["installationAppType"] = _type
        if changelog_destination:
            json["changelogDestination"] = changelog_destination
        if port_app_config:
            json["config"] = port_app_config.to_request()

        response = await self.client.patch(
            f"{self.auth.api_url}/integration/{self.integration_identifier}",
            headers=headers,
            json=json,
        )
        handle_status_code(response)

    async def initialize_integration(
        self,
        _type: str,
        changelog_destination: dict[str, Any],
        port_app_config: Optional["PortAppConfig"] = None,
    ) -> None:
        logger.info(f"Initiating integration with id: {self.integration_identifier}")
        response = await self._get_current_integration()
        if response.status_code == status.HTTP_404_NOT_FOUND:
            await self.create_integration(_type, changelog_destination, port_app_config)
        else:
            handle_status_code(response)

            integration = response.json()["integration"]
            logger.info("Checking for diff in integration configuration")
            if (
                integration["changelogDestination"] != changelog_destination
                or integration["installationAppType"] != _type
            ):
                await self.patch_integration(
                    _type, changelog_destination, port_app_config
                )

        logger.info(
            f"Integration with id: {self.integration_identifier} successfully registered"
        )



import asyncio
import json
import signal
import sys
from typing import Any, Callable, Awaitable

from confluent_kafka import Consumer, KafkaException, Message  # type: ignore
from loguru import logger
from pydantic import BaseModel

from port_ocean.consumers.base_consumer import BaseConsumer


class KafkaConsumerConfig(BaseModel):
    brokers: str
    username: str | None = None
    password: str | None = None
    group_name: str | None = None
    security_protocol: str
    authentication_mechanism: str
    kafka_security_enabled: bool
    consumer_poll_timeout: int


class KafkaConsumer(BaseConsumer):
    def __init__(
        self,
        msg_process: Callable[[dict[Any, Any], str], Awaitable[None]],
        config: KafkaConsumerConfig,
        org_id: str,
    ) -> None:
        self.running = False
        self.org_id = org_id
        self.config = config

        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

        self.msg_process = msg_process
        if config.kafka_security_enabled:
            kafka_config = {
                "bootstrap.servers": config.brokers,
                "security.protocol": config.security_protocol,
                "sasl.mechanism": config.authentication_mechanism,
                "sasl.username": config.username,
                "sasl.password": config.password,
                "group.id": f"{self.org_id}.{config.group_name}",
                "enable.auto.commit": "false",
            }
        else:
            kafka_config = {
                "bootstrap.servers": config.brokers,
                "group.id": "no-security",
                "enable.auto.commit": "false",
            }

        self.consumer = Consumer(kafka_config)

    def _handle_message(self, raw_msg: Message) -> None:
        message = json.loads(raw_msg.value().decode())
        topic = raw_msg.topic()

        async def try_wrapper() -> None:
            try:
                await self.msg_process(message, topic)
            except Exception as e:
                _type, _, tb = sys.exc_info()
                logger.opt(exception=(_type, None, tb)).error(
                    f"Failed to process message: {str(e)}"
                )

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(try_wrapper())

    def start(self) -> None:
        try:
            logger.info("Start consumer...")

            self.consumer.subscribe(
                [f"{self.org_id}.change.log"],
                on_assign=lambda _, partitions: logger.info(
                    f"Assignment: {partitions}"
                ),
            )
            logger.info("Subscribed to topics")
            self.running = True
            while self.running:
                try:
                    msg = self.consumer.poll(timeout=self.config.consumer_poll_timeout)
                    if msg is None:
                        continue
                    if msg.error():
                        raise KafkaException(msg.error())
                    else:
                        try:
                            logger.info(
                                "Process message "
                                f"from topic {msg.topic()}, partition {msg.partition()}, offset {msg.offset()}"
                            )
                            self._handle_message(msg)
                        except Exception as process_error:
                            logger.exception(
                                "Failed process message"
                                f" from topic {msg.topic()}, partition {msg.partition()}, offset {msg.offset()}: {str(process_error)}"
                            )
                        finally:
                            self.consumer.commit(asynchronous=False)
                except Exception as message_error:
                    logger.error(str(message_error))
        finally:
            self.exit_gracefully()

    def exit_gracefully(self, *_: Any) -> None:
        logger.info("Exiting gracefully...")
        self.running = False
        self.consumer.close()



import asyncio
import json
from pathlib import Path
from typing import Type, Any, TypedDict, Optional

import httpx
import yaml
from loguru import logger
from pydantic import BaseModel, Field
from starlette import status

from port_ocean.clients.port.client import PortClient
from port_ocean.config.settings import IntegrationConfiguration
from port_ocean.context.ocean import ocean
from port_ocean.core.handlers.port_app_config.models import PortAppConfig
from port_ocean.exceptions.port_defaults import (
    AbortDefaultCreationError,
    UnsupportedDefaultFileType,
)

YAML_EXTENSIONS = [".yaml", ".yml"]
ALLOWED_FILE_TYPES = [".json", *YAML_EXTENSIONS]


class Preset(TypedDict):
    blueprint: str
    data: list[dict[str, Any]]


class Defaults(BaseModel):
    blueprints: list[dict[str, Any]] = []
    actions: list[Preset] = []
    scorecards: list[Preset] = []
    port_app_config: Optional[PortAppConfig] = Field(
        default=None, alias="port-app-config"
    )

    class Config:
        allow_population_by_field_name = True


async def _is_integration_exists(port_client: PortClient) -> bool:
    try:
        await port_client.get_current_integration(should_log=False)
        return True
    except httpx.HTTPStatusError as e:
        if e.response.status_code != status.HTTP_404_NOT_FOUND:
            raise e

    return False


def deconstruct_blueprints_to_creation_steps(
    raw_blueprints: list[dict[str, Any]]
) -> tuple[list[dict[str, Any]], ...]:
    """
    Deconstructing the blueprint into stages so the api wont fail to create a blueprint if there is a conflict
    example: Preventing the failure of creating a blueprint with a relation to another blueprint
    """
    (
        bare_blueprint,
        with_relations,
        full_blueprint,
    ) = ([], [], [])

    for blueprint in raw_blueprints.copy():
        full_blueprint.append(blueprint.copy())

        blueprint.pop("calculationProperties", {})
        blueprint.pop("mirrorProperties", {})
        with_relations.append(blueprint.copy())

        blueprint.pop("teamInheritance", {})
        blueprint.pop("relations", {})
        bare_blueprint.append(blueprint)

    return (
        bare_blueprint,
        with_relations,
        full_blueprint,
    )


async def _create_resources(
    port_client: PortClient,
    defaults: Defaults,
    integration_config: IntegrationConfiguration,
) -> None:
    creation_stage, *blueprint_patches = deconstruct_blueprints_to_creation_steps(
        defaults.blueprints
    )

    create_results = await asyncio.gather(
        *(port_client.create_blueprint(blueprint) for blueprint in creation_stage),
        return_exceptions=True,
    )

    errors = [result for result in create_results if isinstance(result, Exception)]
    created_blueprints = [
        result["identifier"]
        for result in create_results
        if not isinstance(result, Exception)
    ]

    if errors:
        for error in errors:
            if isinstance(error, httpx.HTTPStatusError):
                logger.warning(
                    f"Failed to create resources: {error.response.text}. Rolling back changes..."
                )

        raise AbortDefaultCreationError(created_blueprints, errors)

    try:
        for patch_stage in blueprint_patches:
            await asyncio.gather(
                *(
                    port_client.patch_blueprint(blueprint["identifier"], blueprint)
                    for blueprint in patch_stage
                )
            )

        await asyncio.gather(
            *(
                port_client.create_action(blueprint_actions["blueprint"], action)
                for blueprint_actions in defaults.actions
                for action in blueprint_actions["data"]
            )
        )

        await asyncio.gather(
            *(
                port_client.create_scorecard(blueprint_scorecards["blueprint"], action)
                for blueprint_scorecards in defaults.scorecards
                for action in blueprint_scorecards["data"]
            )
        )

        await port_client.initialize_integration(
            integration_config.integration.type,
            integration_config.event_listener.to_request(),
            port_app_config=defaults.port_app_config,
        )
    except httpx.HTTPStatusError as e:
        logger.error(
            f"Failed to create resources: {e.response.text}. Rolling back changes..."
        )
        raise AbortDefaultCreationError(created_blueprints, [e])


async def _initialize_defaults(
    config_class: Type[PortAppConfig], integration_config: IntegrationConfiguration
) -> None:
    port_client = ocean.port_client
    is_exists = await _is_integration_exists(port_client)
    if is_exists:
        return None
    defaults = get_port_integration_defaults(config_class)
    if not defaults:
        return None

    try:
        await _create_resources(port_client, defaults, integration_config)
    except AbortDefaultCreationError as e:
        logger.warning(
            f"Failed to create resources. Rolling back blueprints : {e.blueprints_to_rollback}"
        )
        await asyncio.gather(
            *(
                port_client.delete_blueprint(
                    blueprint["identifier"], should_raise=False
                )
                for blueprint in defaults.blueprints
            )
        )

        raise ExceptionGroup(str(e), e.errors)


def initialize_defaults(
    config_class: Type[PortAppConfig], integration_config: IntegrationConfiguration
) -> None:
    try:
        asyncio.new_event_loop().run_until_complete(
            _initialize_defaults(config_class, integration_config)
        )
    except Exception as e:
        logger.debug(f"Failed to initialize defaults, skipping... Error: {e}")


def get_port_integration_defaults(
    port_app_config_class: Type[PortAppConfig], base_path: Path = Path(".")
) -> Defaults | None:
    defaults_dir = base_path / ".port/resources"
    if not defaults_dir.exists():
        return None

    if not defaults_dir.is_dir():
        raise UnsupportedDefaultFileType(
            f"Defaults directory is not a directory: {defaults_dir}"
        )

    default_jsons = {}
    allowed_file_names = [
        field_model.alias for _, field_model in Defaults.__fields__.items()
    ]
    for path in defaults_dir.iterdir():
        if path.stem in allowed_file_names:
            if not path.is_file() or path.suffix not in ALLOWED_FILE_TYPES:
                raise UnsupportedDefaultFileType(
                    f"Defaults directory should contain only one of the next types: {ALLOWED_FILE_TYPES}. Found: {path}"
                )

            if path.suffix in YAML_EXTENSIONS:
                default_jsons[path.stem] = yaml.safe_load(path.read_text())
            else:
                default_jsons[path.stem] = json.loads(path.read_text())

    return Defaults(
        blueprints=default_jsons.get("blueprints", []),
        actions=default_jsons.get("actions", []),
        scorecards=default_jsons.get("scorecards", []),
        port_app_config=port_app_config_class(
            **default_jsons.get("port-app-config", {})
        ),
    )



from port_ocean.exceptions.base import BaseOceanException


class OceanAbortException(BaseOceanException):
    pass


class RawObjectValidationException(OceanAbortException):
    pass


class EntityProcessorException(BaseOceanException):
    pass


class RelationValidationException(OceanAbortException):
    pass


class UnsupportedEventListenerTypeException(BaseOceanException):
    pass


class IntegrationAlreadyStartedException(BaseOceanException):
    pass


class IntegrationNotStartedException(BaseOceanException):
    pass



import sys

import loguru
from loguru import logger

from port_ocean.config.settings import LogLevelType


def setup_logger(level: LogLevelType) -> None:
    logger_format = (
        "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
        "<level>{level: <8}</level> | "
        "<level>{message}</level>"
    )
    if level == "DEBUG":
        logger_format += " | {extra}"

    logger.remove()
    logger.add(
        sys.stdout,
        level=level.upper(),
        format=logger_format,
        enqueue=True,  # process logs in background
        diagnose=False,  # hide variable values in log backtrace
    )
    logger.configure(patcher=exception_deserializer)


def exception_deserializer(record: "loguru.Record") -> None:
    """
    Workaround for when trying to log exception objects with loguru.
    loguru doesn't able to deserialize `Exception` subclasses.
    https://github.com/Delgan/loguru/issues/504#issuecomment-917365972
    """
    exception: loguru.RecordException | None = record["exception"]
    if exception is not None:
        fixed = Exception(str(exception.value))
        record["exception"] = exception._replace(value=fixed)



import sys
from typing import Callable

from fastapi import FastAPI, APIRouter
from loguru import logger
from pydantic import BaseModel
from starlette.types import Scope, Receive, Send

from port_ocean.clients.port.client import PortClient
from port_ocean.config.settings import (
    IntegrationConfiguration,
)
from port_ocean.context.ocean import (
    PortOceanContext,
    ocean,
    initialize_port_ocean_context,
)
from port_ocean.core.integrations.base import BaseIntegration
from port_ocean.middlewares import request_handler


class Ocean:
    def __init__(
        self,
        app: FastAPI | None = None,
        integration_class: Callable[[PortOceanContext], BaseIntegration] | None = None,
        integration_router: APIRouter | None = None,
        config_factory: Callable[..., BaseModel] | None = None,
    ):
        initialize_port_ocean_context(self)
        self.fast_api_app = app or FastAPI()
        self.fast_api_app.middleware("http")(request_handler)

        self.config = IntegrationConfiguration(base_path="./")
        if config_factory:
            self.config.integration.config = config_factory(
                **self.config.integration.config
            ).dict()
        self.integration_router = integration_router or APIRouter()

        self.port_client = PortClient(
            base_url=self.config.port.base_url,
            client_id=self.config.port.client_id,
            client_secret=self.config.port.client_secret,
            integration_identifier=self.config.integration.identifier,
            integration_type=self.config.integration.type,
        )
        self.integration = (
            integration_class(ocean) if integration_class else BaseIntegration(ocean)
        )

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        self.fast_api_app.include_router(self.integration_router, prefix="/integration")

        @self.fast_api_app.on_event("startup")
        async def startup() -> None:
            try:
                await self.integration.start()
            except Exception as e:
                logger.error(f"Failed to start integration with error: {e}")
                sys.exit("Server stopped")

        await self.fast_api_app(scope, receive, send)



from typing import Any

from loguru import logger

from port_ocean.core.handlers.port_app_config.base import BasePortAppConfig


class APIPortAppConfig(BasePortAppConfig):
    async def _get_port_app_config(self) -> dict[str, Any]:
        logger.info("Fetching port app config")
        integration = await self.context.port_client.get_current_integration()
        return integration["config"]



import os
import re
from pathlib import Path
from typing import Any

import yaml
from humps import decamelize
from pydantic import BaseSettings
from pydantic.env_settings import EnvSettingsSource, InitSettingsSource

PROVIDER_WRAPPER_PATTERN = r"{{ from (.*) }}"
PROVIDER_CONFIG_PATTERN = r"^[a-zA-Z0-9]+ .*$"


def read_yaml_config_settings_source(
    settings: "BaseOceanSettings", base_path: str
) -> str:
    yaml_file = getattr(settings.__config__, "yaml_file", "")

    assert yaml_file, "Settings.yaml_file not properly configured"

    path = Path(base_path, yaml_file)

    if not path.exists():
        raise FileNotFoundError(f"Could not open yaml settings file at: {path}")

    return path.read_text("utf-8")


def parse_config_provider(value: str) -> tuple[str, str]:
    match = re.match(PROVIDER_CONFIG_PATTERN, value)
    if not match:
        raise ValueError(
            f"Invalid pattern: {value}. Pattern should match: {PROVIDER_CONFIG_PATTERN}"
        )

    provider_type, provider_value = value.split(" ", 1)

    return provider_type, provider_value


def load_from_config_provider(config_provider: str) -> Any:
    provider_type, value = parse_config_provider(config_provider)
    if provider_type == "env":
        result = os.environ.get(value)
        if result is None:
            raise ValueError(f"Environment variable not found: {value}")
        return result
    else:
        raise ValueError(f"Invalid provider type: {provider_type}")


def decamelize_object(obj: Any) -> Any:
    if isinstance(obj, dict):
        return {decamelize(k): decamelize_object(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [decamelize_object(v) for v in obj]
    else:
        return obj


def parse_config(
    config: dict[str, Any], existing_data: dict[str, Any]
) -> dict[str, Any]:
    """
    Normalizing the config yaml file to work with snake_case and getting only the data that is missing for the settings
    """
    for key, value in config.items():
        decamelize_key = decamelize(key)
        if isinstance(value, dict):
            existing_data[decamelize_key] = parse_config(
                value, existing_data.get(decamelize_key, {})
            )
        elif isinstance(value, str):
            # If the value is a provider, we try to load it from the provider
            if provider_match := re.match(PROVIDER_WRAPPER_PATTERN, value):
                # If the there is already value for that field, we ignore it
                # If the provider failed to load, we ignore it
                if decamelize_key not in existing_data:
                    try:
                        existing_data[decamelize_key] = load_from_config_provider(
                            provider_match.group(1)
                        )
                    except ValueError:
                        pass
            else:
                existing_data[decamelize_key] = value
        else:
            existing_data[decamelize_key] = value
    return existing_data


def load_providers(
    settings: "BaseOceanSettings", existing_values: dict[str, Any], base_path: str
) -> dict[str, Any]:
    yaml_content = read_yaml_config_settings_source(settings, base_path)
    data = yaml.safe_load(yaml_content)
    return parse_config(data, existing_values)


class BaseOceanSettings(BaseSettings):
    base_path: str

    class Config:
        yaml_file = "./config.yaml"
        env_prefix = "OCEAN__"
        env_nested_delimiter = "__"
        env_file = ".env"
        env_file_encoding = "utf-8"

        @classmethod
        def customise_sources(  # type: ignore
            cls,
            init_settings: InitSettingsSource,
            env_settings: EnvSettingsSource,
            *_,
            **__,
        ):
            return (
                env_settings,
                lambda s: load_providers(
                    s, env_settings(s), init_settings.init_kwargs["base_path"]
                ),
                init_settings,
            )



from abc import abstractmethod
from dataclasses import dataclass, field

from loguru import logger

from port_ocean.core.base import BaseWithContext
from port_ocean.core.handlers.port_app_config.models import ResourceConfig
from port_ocean.core.models import Entity
from port_ocean.core.ocean_types import RawEntityDiff, EntityDiff


@dataclass
class EntityPortDiff:
    deleted: list[Entity] = field(default_factory=list)
    modified: list[Entity] = field(default_factory=list)
    created: list[Entity] = field(default_factory=list)


class BaseEntityProcessor(BaseWithContext):
    @abstractmethod
    async def _parse_items(
        self, mapping: ResourceConfig, raw_data: RawEntityDiff
    ) -> EntityDiff:
        pass

    async def parse_items(
        self, mapping: ResourceConfig, raw_data: RawEntityDiff
    ) -> EntityDiff:
        with logger.contextualize(kind=mapping.kind):
            return await self._parse_items(mapping, raw_data)



import threading
from typing import Any, Callable, Literal

from loguru import logger

from port_ocean.consumers.kafka_consumer import KafkaConsumer, KafkaConsumerConfig
from port_ocean.context.ocean import (
    PortOceanContext,
    initialize_port_ocean_context,
    ocean,
)
from port_ocean.core.event_listener.base import (
    BaseEventListener,
    EventListenerEvents,
    EventListenerSettings,
)


class KafkaEventListenerSettings(EventListenerSettings):
    type: Literal["KAFKA"]
    brokers: str = "b-1-public.publicclusterprod.t9rw6w.c1.kafka.eu-west-1.amazonaws.com:9196,b-2-public.publicclusterprod.t9rw6w.c1.kafka.eu-west-1.amazonaws.com:9196,b-3-public.publicclusterprod.t9rw6w.c1.kafka.eu-west-1.amazonaws.com:9196"
    security_protocol: str = "SASL_SSL"
    authentication_mechanism: str = "SCRAM-SHA-512"
    kafka_security_enabled: bool = True
    consumer_poll_timeout: int = 1


class KafkaEventListener(BaseEventListener):
    def __init__(
        self,
        events: EventListenerEvents,
        event_listener_config: KafkaEventListenerSettings,
        org_id: str,
        integration_identifier: str,
        integration_type: str,
    ):
        super().__init__(events)
        self.event_listener_config = event_listener_config
        self.org_id = org_id
        self.integration_identifier = integration_identifier
        self.integration_type = integration_type

    async def _get_kafka_config(self) -> KafkaConsumerConfig:
        if self.event_listener_config.kafka_security_enabled:
            creds = await ocean.port_client.get_kafka_creds()
            return KafkaConsumerConfig(
                **self.event_listener_config.dict(),
                username=creds.get("username"),
                password=creds.get("password"),
                group_name=f"{self.integration_type}.{self.integration_identifier}",
            )

        return KafkaConsumerConfig.parse_obj(self.event_listener_config.dict())

    def should_be_processed(self, msg_value: dict[Any, Any], topic: str) -> bool:
        integration_identifier = (
            msg_value.get("diff", {}).get("after", {}).get("identifier")
        )
        if integration_identifier == self.integration_identifier and (
            "change.log" in topic
        ):
            return msg_value.get("changelogDestination", {}).get("type", "") == "KAFKA"

        return False

    async def _handle_message(self, message: dict[Any, Any], topic: str) -> None:
        if not self.should_be_processed(message, topic):
            return

        if "change.log" in topic and message is not None:
            await self.events["on_resync"](message)

    def wrapped_start(
        self, context: PortOceanContext, func: Callable[[], None]
    ) -> Callable[[], None]:
        ocean_app = context.app

        def wrapper() -> None:
            initialize_port_ocean_context(ocean_app=ocean_app)
            func()

        return wrapper

    async def start(self) -> None:
        consumer = KafkaConsumer(
            msg_process=self._handle_message,
            config=await self._get_kafka_config(),
            org_id=self.org_id,
        )
        logger.info("Starting Kafka consumer")
        threading.Thread(
            name="ocean_kafka_consumer",
            target=self.wrapped_start(ocean, consumer.start),
        ).start()



import asyncio
from collections import defaultdict
from itertools import groupby

from port_ocean.clients.port.client import PortClient
from port_ocean.core.models import Entity


async def get_related_entities(
    entities: list[Entity], port_client: PortClient
) -> list[Entity]:
    entities_with_relations = [entity for entity in entities if entity.relations]
    blueprint_identifier_to_entity = dict(
        groupby(
            entities_with_relations,
            key=lambda x: x.blueprint,
        )
    )
    blueprints = await asyncio.gather(
        *(
            port_client.get_blueprint(blueprint_identifier)
            for blueprint_identifier in blueprint_identifier_to_entity.keys()
        )
    )
    entity_to_blueprint = [
        (
            entity,
            next(
                blueprint
                for blueprint in blueprints
                if blueprint.identifier == entity.blueprint
            ),
        )
        for entity in entities_with_relations
    ]

    blueprints_to_relations = defaultdict(list)
    for entity, blueprint in entity_to_blueprint:
        for relation_name, relation in entity.relations.items():
            relation_blueprint = blueprint.relations[relation_name].target
            blueprints_to_relations[relation_blueprint].extend(
                relation if isinstance(relation, list) else [relation]
            )

    return [
        Entity(identifier=relation, blueprint=blueprint)
        for blueprint, relations in blueprints_to_relations.items()
        # multiple entities can point to the same relation in the same blueprint, for performance reasons
        # we want to avoid fetching the same relation multiple times
        for relation in set(relations)
    ]



from collections import defaultdict

from port_ocean.core.ocean_types import (
    IntegrationEventsCallbacks,
    START_EVENT_LISTENER,
    RESYNC_EVENT_LISTENER,
)


class EventsMixin:
    def __init__(self) -> None:
        self.event_strategy: IntegrationEventsCallbacks = {
            "start": [],
            "resync": defaultdict(list),
        }

    def on_start(self, func: START_EVENT_LISTENER) -> START_EVENT_LISTENER:
        self.event_strategy["start"].append(func)
        return func

    def on_resync(
        self, func: RESYNC_EVENT_LISTENER, kind: str | None = None
    ) -> RESYNC_EVENT_LISTENER:
        self.event_strategy["resync"][kind].append(func)
        return func



import os
import shutil
from io import BytesIO

import click
import httpx

from port_ocean.cli.commands.main import cli_start, console


def download_github_folder(
    owner: str, repo_name: str, folder_path: str, destination_path: str
) -> None:
    # Construct the API URL to get the contents of the folder
    api_url = f"https://api.github.com/repos/{owner}/{repo_name}/contents/{folder_path}"

    # Send a GET request to the API
    response = httpx.get(api_url)

    # Check if the request was successful
    if response.is_error:
        console.print(
            f"[bold red]Failed to download the folder `{folder_path}`.[/bold red] Status Code: {response.status_code}, Error: {response.text}"
        )
        exit(1)

    # Create the destination folder if it doesn't exist
    if not os.path.exists(destination_path):
        os.makedirs(destination_path)

    # Iterate over the files and download them
    repo_contents = response.json()
    for content in repo_contents:
        if content["type"] == "file":
            file_url = content["download_url"]
            file_name = os.path.join(destination_path, content["name"])

            # Download the file
            with httpx.stream("GET", file_url) as file_response:
                if file_response.status_code == 200:
                    with open(file_name, "wb") as file:
                        shutil.copyfileobj(BytesIO(file_response.content), file)
                else:
                    console.print(
                        f"[bold red]Failed to download file `{content['name']}`.[/bold red] Status code: {file_response.status_code}, Error: {file_response.text}"
                    )
                    exit(1)

    console.print(f"Folder `{folder_path}` downloaded successfully!")


@cli_start.command()
@click.argument("name", type=str)
@click.option(
    "-p",
    "--path",
    "path",
    default=None,
    type=click.Path(exists=True),
    help="Desired path to pull the integration to. defaults to ./NAME",
)
def pull(name: str, path: str) -> None:
    """
    Pull an integration bt the NAME from the list of available public integrations.

    NAME: Name of the integration to pull.
    """
    download_github_folder(
        "port-labs", "Port-Ocean", f"integrations/{name}", path or f"./{name}"
    )



import abc

from fastapi.responses import Response, PlainTextResponse
from port_ocean.exceptions.base import BaseOceanException


class BaseAPIException(BaseOceanException, abc.ABC):
    @abc.abstractmethod
    def response(self) -> Response:
        pass


class InternalServerException(BaseAPIException):
    def response(self) -> Response:
        return PlainTextResponse(content="Internal server error", status_code=500)



from abc import abstractmethod
from typing import TypedDict, Callable, Any, Awaitable

from pydantic import BaseModel, Extra


class EventListenerEvents(TypedDict):
    on_resync: Callable[[dict[Any, Any]], Awaitable[None]]


class BaseEventListener:
    def __init__(
        self,
        events: EventListenerEvents,
    ):
        self.events = events

    @abstractmethod
    async def start(self) -> None:
        pass


class EventListenerSettings(BaseModel, extra=Extra.allow):
    type: str

    def to_request(self) -> dict[str, Any]:
        return {"type": self.type}



from typing import Optional, Any

from pydantic import BaseModel, Field

from port_ocean.clients.port.types import RequestOptions


class EntityMapping(BaseModel):
    identifier: str
    title: str | None
    blueprint: str
    properties: dict[str, str] = Field(default_factory=dict)
    relations: dict[str, str] = Field(default_factory=dict)


class PortResourceConfig(BaseModel):
    class MappingsConfig(BaseModel):
        mappings: EntityMapping

    entity: Optional[MappingsConfig]


class ResourceConfig(BaseModel):
    class Selector(BaseModel):
        query: str

    kind: str
    selector: Selector
    port: PortResourceConfig


class PortAppConfig(BaseModel):
    enable_merge_entity: bool = Field(alias="enableMergeEntity", default=False)
    delete_dependent_entities: bool = Field(
        alias="deleteDependentEntities", default=False
    )
    create_missing_related_entities: bool = Field(
        alias="createMissingRelatedEntities", default=False
    )
    resources: list[ResourceConfig] = Field(default_factory=list)

    def get_port_request_options(self) -> RequestOptions:
        return {
            "delete_dependent_entities": self.delete_dependent_entities,
            "create_missing_related_entities": self.create_missing_related_entities,
            "merge": self.enable_merge_entity,
        }

    def to_request(self) -> dict[str, Any]:
        return {
            "deleteDependentEntities": self.delete_dependent_entities,
            "createMissingRelatedEntities": self.create_missing_related_entities,
            "enableMergeEntity": self.enable_merge_entity,
            "resources": [resource.dict(by_alias=True) for resource in self.resources],
        }

    class Config:
        allow_population_by_field_name = True
        validate_assignment = True



from .http.event_listener import HttpEventListener, HttpEventListenerSettings
from .kafka.event_listener import KafkaEventListener, KafkaEventListenerSettings
from .polling.event_listener import PollingEventListener, PollingEventListenerSettings

EventListenerSettingsType = (
    HttpEventListenerSettings
    | KafkaEventListenerSettings
    | PollingEventListenerSettings
)

__all__ = [
    "EventListenerSettingsType",
    "HttpEventListener",
    "HttpEventListenerSettings",
    "KafkaEventListener",
    "KafkaEventListenerSettings",
    "PollingEventListener",
    "PollingEventListenerSettings",
]



from typing import Literal, Any

from fastapi import APIRouter
from loguru import logger
from pydantic import AnyHttpUrl

from port_ocean.context.ocean import ocean
from port_ocean.core.event_listener.base import (
    BaseEventListener,
    EventListenerEvents,
    EventListenerSettings,
)


class HttpEventListenerSettings(EventListenerSettings):
    type: Literal["WEBHOOK"]
    app_host: AnyHttpUrl

    def to_request(self) -> dict[str, Any]:
        return {
            **super().to_request(),
            "url": self.app_host + "/resync",
        }


class HttpEventListener(BaseEventListener):
    def __init__(
        self,
        events: EventListenerEvents,
        event_listener_config: HttpEventListenerSettings,
    ):
        super().__init__(events)
        self.event_listener_config = event_listener_config

    async def start(self) -> None:
        logger.info("Setting up HTTP Event Listener")
        target_channel_router = APIRouter()

        @target_channel_router.post("/resync")
        async def resync() -> None:
            await self.events["on_resync"]({})

        ocean.app.fast_api_app.include_router(target_channel_router)



from urllib.parse import quote_plus

import httpx
from loguru import logger

from port_ocean.clients.port.authentication import PortAuthentication
from port_ocean.clients.port.types import RequestOptions, UserAgentType
from port_ocean.clients.utils import handle_status_code
from port_ocean.core.models import Entity


class EntityClientMixin:
    def __init__(self, auth: PortAuthentication, client: httpx.AsyncClient):
        self.auth = auth
        self.client = client

    async def upsert_entity(
        self,
        entity: Entity,
        request_options: RequestOptions,
        user_agent_type: UserAgentType | None = None,
        should_raise: bool = True,
    ) -> None:
        validation_only = request_options.get("validation_only", False)
        logger.info(
            f"{'Validating' if validation_only else 'Upserting'} entity: {entity.identifier} of blueprint: {entity.blueprint}"
        )
        headers = await self.auth.headers(user_agent_type)

        response = await self.client.post(
            f"{self.auth.api_url}/blueprints/{entity.blueprint}/entities",
            json=entity.dict(exclude_unset=True, by_alias=True),
            headers=headers,
            params={
                "upsert": "true",
                "merge": str(request_options.get("merge", False)).lower(),
                "create_missing_related_entities": str(
                    request_options.get("create_missing_related_entities", False)
                ).lower(),
                "validation_only": str(validation_only).lower(),
            },
        )

        if response.is_error:
            logger.error(
                f"Error {'Validating' if validation_only else 'Upserting'} "
                f"entity: {entity.identifier} of "
                f"blueprint: {entity.blueprint}"
            )
        handle_status_code(response, should_raise)

    async def delete_entity(
        self,
        entity: Entity,
        request_options: RequestOptions,
        user_agent_type: UserAgentType | None = None,
        should_raise: bool = True,
    ) -> None:
        logger.info(
            f"Delete entity: {entity.identifier} of blueprint: {entity.blueprint}"
        )
        response = await self.client.delete(
            f"{self.auth.api_url}/blueprints/{entity.blueprint}/entities/{quote_plus(entity.identifier)}",
            headers=await self.auth.headers(user_agent_type),
            params={
                "delete_dependents": str(
                    request_options.get("delete_dependent_entities", False)
                ).lower()
            },
        )

        if response.is_error:
            logger.error(
                f"Error deleting "
                f"entity: {entity.identifier} of "
                f"blueprint: {entity.blueprint}"
            )

        handle_status_code(response, should_raise)

    async def validate_entity_exist(self, identifier: str, blueprint: str) -> None:
        logger.info(f"Validating entity {identifier} of blueprint {blueprint} exists")

        response = await self.client.get(
            f"{self.auth.api_url}/blueprints/{blueprint}/entities/{identifier}",
            headers=await self.auth.headers(),
        )
        if response.is_error:
            logger.error(
                f"Error validating "
                f"entity: {identifier} of "
                f"blueprint: {blueprint}"
            )
        handle_status_code(response)

    async def search_entities(self, user_agent_type: UserAgentType) -> list[Entity]:
        query = {
            "combinator": "and",
            "rules": [
                {
                    "property": "$datasource",
                    "operator": "=",
                    "value": self.auth.user_agent(user_agent_type),
                },
            ],
        }

        logger.info(f"Searching entities with query {query}")
        response = await self.client.post(
            f"{self.auth.api_url}/entities/search",
            json=query,
            headers=await self.auth.headers(user_agent_type),
            params={
                "exclude_calculated_properties": "true",
                "include": ["blueprint", "identifier"],
            },
        )
        handle_status_code(response)
        return [Entity.parse_obj(result) for result in response.json()["entities"]]

    async def search_dependent_entities(self, entity: Entity) -> list[Entity]:
        body = {
            "combinator": "and",
            "rules": [
                {
                    "operator": "relatedTo",
                    "blueprint": entity.blueprint,
                    "value": entity.identifier,
                    "direction": "downstream",
                }
            ],
        }

        logger.info(f"Searching dependent entity with body {body}")
        response = await self.client.post(
            f"{self.auth.api_url}/entities/search",
            headers=await self.auth.headers(),
            json=body,
        )
        handle_status_code(response)

        return [Entity.parse_obj(result) for result in response.json()["entities"]]

    async def validate_entity_payload(
        self, entity: Entity, options: RequestOptions
    ) -> None:
        logger.info(f"Validating entity {entity.identifier}")
        await self.upsert_entity(
            entity,
            {
                "merge": options.get("merge", False),
                "create_missing_related_entities": options.get(
                    "create_missing_related_entities", False
                ),
                "validation_only": True,
            },
        )



from importlib.util import find_spec

try:
    find_spec("click")
    find_spec("rich")
    find_spec("cookiecutter")
    find_spec("jinja2-time")

    cli_included = True
except ImportError:
    raise ImportError(
        "The CLI is not available. Please install the CLI dependencies using: pip install port-ocean[cli]"
    )



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from port_ocean.context.ocean import PortOceanContext


class BaseWithContext:
    def __init__(self, context: "PortOceanContext"):
        self.context = context



import asyncio

from port_ocean.clients.port.client import PortClient
from port_ocean.core.handlers.entities_state_applier.port.get_related_entities import (
    get_related_entities,
)
from port_ocean.core.handlers.entity_processor.base import EntityPortDiff
from port_ocean.core.utils import is_same_entity
from port_ocean.exceptions.core import RelationValidationException


async def validate_entity_relations(
    diff: EntityPortDiff, port_client: PortClient
) -> None:
    modified_or_created_entities = diff.modified + diff.created
    related_entities = await get_related_entities(
        modified_or_created_entities, port_client
    )

    required_entities = []

    for entity in related_entities:
        if any(is_same_entity(item, entity) for item in diff.deleted):
            raise RelationValidationException(
                f"Cant delete entity {entity} of blueprint {entity.blueprint} "
                f"because it was specified as relation target of entity {entity} "
                f"of blueprint {entity.blueprint}"
            )

        if not any(
            is_same_entity(item, entity) for item in modified_or_created_entities
        ):
            required_entities.append(entity)

    await asyncio.gather(
        *(
            port_client.validate_entity_exist(item.identifier, item.blueprint)
            for item in required_entities
        )
    )



from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from typing import AsyncIterator, Literal, Any, TYPE_CHECKING, Optional
from uuid import uuid4

from loguru import logger
from werkzeug.local import LocalStack, LocalProxy

from port_ocean.exceptions.context import EventContextNotFoundError
from port_ocean.utils import get_time

if TYPE_CHECKING:
    from port_ocean.core.handlers.port_app_config.models import (
        PortAppConfig,
        ResourceConfig,
    )

TriggerType = Literal["manual", "machine", "request"]


class EventType:
    START = "start"
    RESYNC = "resync"
    HTTP_REQUEST = "http_request"


@dataclass
class EventContext:
    event_type: str
    trigger_type: TriggerType = "machine"
    resource_config: Optional["ResourceConfig"] = None
    attributes: dict[str, Any] = field(default_factory=dict)
    _port_app_config: Optional["PortAppConfig"] = None
    _parent_event: Optional["EventContext"] = None
    _event_id: str = field(default_factory=lambda: str(uuid4()))

    @property
    def id(self) -> str:
        return self._event_id

    @property
    def parent(self) -> Optional["EventContext"]:
        return self._parent_event

    @property
    def parent_id(self) -> Optional[str]:
        return self._parent_event.id if self._parent_event else None

    @property
    def port_app_config(self) -> "PortAppConfig":
        if self._port_app_config is None:
            raise ValueError("Port app config is not set")
        return self._port_app_config

    @port_app_config.setter
    def port_app_config(self, value: "PortAppConfig") -> None:
        self._port_app_config = value


_event_context_stack: LocalStack[EventContext] = LocalStack()


def _get_event_context() -> EventContext:
    """
    Get the event context from the current thread.
    """
    top_event_context = _event_context_stack.top
    if top_event_context is None:
        raise EventContextNotFoundError(
            "You must be inside an event context in order to use it"
        )

    return top_event_context


event: EventContext = LocalProxy(lambda: _get_event_context())  # type: ignore


@asynccontextmanager
async def event_context(
    event_type: str,
    trigger_type: TriggerType = "manual",
    resource_config: Optional["ResourceConfig"] = None,
    attributes: dict[str, Any] | None = None,
) -> AsyncIterator[EventContext]:
    attributes = attributes or {}

    parent = _event_context_stack.top

    _event_context_stack.push(
        EventContext(
            event_type,
            trigger_type=trigger_type,
            resource_config=resource_config,
            attributes=attributes,
            _parent_event=parent,
            # inherit port app config from parent event so it can be used in nested events
            _port_app_config=parent.port_app_config if parent else None,
        )
    )

    start_time = get_time(seconds_precision=False)
    with logger.contextualize(
        event_trigger_type=event.trigger_type,
        event_kind=event.event_type,
        event_id=event.id,
        event_parent_id=event.parent_id,
        event_resource_kind=event.resource_config.kind
        if event.resource_config
        else None,
    ):
        logger.info("Event started")
        try:
            yield event
        except:
            success = False
            raise
        else:
            success = True
        finally:
            end_time = get_time(seconds_precision=False)
            time_elapsed = round(end_time - start_time, 5)
            logger.bind(
                success=success,
                time_elapsed=time_elapsed,
            ).info("Event finished")

    _event_context_stack.pop()



from typing import TypedDict, Any, AsyncIterator, Callable, Awaitable

from port_ocean.core.models import Entity


class RawEntityDiff(TypedDict):
    before: list[dict[Any, Any]]
    after: list[dict[Any, Any]]


class EntityDiff(TypedDict):
    before: list[Entity]
    after: list[Entity]


RAW_ITEM = dict[Any, Any]
RAW_RESULT = list[RAW_ITEM]
ASYNC_GENERATOR_RESYNC_TYPE = AsyncIterator[RAW_RESULT]
RESYNC_RESULT = list[RAW_ITEM | ASYNC_GENERATOR_RESYNC_TYPE]

LISTENER_RESULT = Awaitable[RAW_RESULT] | ASYNC_GENERATOR_RESYNC_TYPE
RESYNC_EVENT_LISTENER = Callable[[str], LISTENER_RESULT]
START_EVENT_LISTENER = Callable[[], Awaitable]


class IntegrationEventsCallbacks(TypedDict):
    start: list[START_EVENT_LISTENER]
    resync: dict[str | None, list[RESYNC_EVENT_LISTENER]]



# -*- coding: utf-8 -*-

import click
from cookiecutter.main import cookiecutter  # type: ignore

from port_ocean import __version__
from port_ocean.cli.commands.main import cli_start, print_logo, console
from port_ocean.cli.utils import cli_root_path


@cli_start.command()
@click.argument("path", default=".", type=click.Path(exists=True))
def new(path: str) -> None:
    """
    Scaffold a new integration in the given PATH.

    PATH: Path to the integration. If not provided, the current directory will be used.
    """
    print_logo()

    console.print(
        "🚢 Unloading cargo... Setting up your integration at the dock.", style="bold"
    )

    result = cookiecutter(
        f"{cli_root_path}/cookiecutter",
        output_dir=path,
        extra_context={"version": __version__},
    )
    name = result.split("/")[-1]

    console.print(
        "\n🌊 Ahoy, Captain! Your project is ready to set sail into the vast ocean of possibilities!",
        style="bold",
    )
    console.print("Here are your next steps:\n", style="bold")
    console.print(
        "⚓️ Install necessary packages: Run [bold][blue]make install[/blue][/bold] to install all required packages for your project.\n"
        f"▶️ [bold][blue]cd {path}/{name} && make install && . .venv/bin/activate[/blue][/bold]\n"
    )
    console.print(
        "⚓️ Set sail with [blue]Ocean[/blue]: Run [bold][blue]ocean sail[/blue] <path_to_integration>[/bold] to run the project using Ocean.\n"
        f"▶️ [bold][blue]ocean sail {path}/{name}[/blue][/bold] \n"
    )
    console.print(
        "⚓️ Smooth sailing with [blue]Make[/blue]: Alternatively, you can run [bold][blue]make run[/blue][/bold] to launch your project using Make. \n"
        f"▶️ [bold][blue]make run {path}/{name}[/blue][/bold]"
    )



import sys
from importlib.util import spec_from_file_location, module_from_spec
from inspect import getmembers, isclass
from types import ModuleType
from typing import Type

import uvicorn

from port_ocean.config.dynamic import default_config_factory
from port_ocean.config.settings import LogLevelType, ApplicationSettings
from port_ocean.core.integrations.base import BaseIntegration
from port_ocean.logger_setup import setup_logger
from port_ocean.ocean import Ocean
from port_ocean.port_defaults import initialize_defaults
from port_ocean.utils import get_spec_file


def _get_base_integration_class_from_module(
    module: ModuleType,
) -> Type[BaseIntegration]:
    for name, obj in getmembers(module):
        if (
            isclass(obj)
            and type(obj) == type
            and issubclass(obj, BaseIntegration)
            and obj != BaseIntegration
        ):
            return obj

    raise Exception(f"Failed to load integration from module: {module.__name__}")


def _load_module(file_path: str) -> ModuleType:
    spec = spec_from_file_location("module_name", file_path)
    if spec is None or spec.loader is None:
        raise Exception(f"Failed to load integration from path: {file_path}")

    module = module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


def _create_default_app(path: str | None = None) -> Ocean:
    sys.path.append(".")
    try:
        integration_path = f"{path}/integration.py" if path else "integration.py"
        module = _load_module(integration_path)
        integration_class = _get_base_integration_class_from_module(module)
    except Exception:
        integration_class = None

    spec = get_spec_file()
    config_factory = None
    if spec is not None:
        config_factory = default_config_factory(spec.get("configurations", []))
    return Ocean(integration_class=integration_class, config_factory=config_factory)


def run(
    path: str = ".",
    log_level: LogLevelType = "DEBUG",
    port: int = 8000,
    initialize_port_resources: bool | None = None,
) -> None:
    application_settings = ApplicationSettings(log_level=log_level, port=port)

    setup_logger(application_settings.log_level)
    default_app = _create_default_app(path)

    main_path = f"{path}/main.py" if path else "main.py"
    app_module = _load_module(main_path)
    app: Ocean = {name: item for name, item in getmembers(app_module)}.get(
        "app", default_app
    )

    # Override config with arguments
    if initialize_port_resources is not None:
        app.config.initialize_port_resources = initialize_port_resources
    if app.config.initialize_port_resources:
        initialize_defaults(
            app.integration.AppConfigHandlerClass.CONFIG_CLASS, app.config
        )

    uvicorn.run(app, host="0.0.0.0", port=application_settings.port)



# -*- coding: utf-8 -*-
import httpx

from port_ocean.cli.commands.main import cli_start, console


def list_git_folders(owner: str, repo_name: str, path: str) -> list[str]:
    # Construct the API URL to get the contents of the folder
    api_url = f"https://api.github.com/repos/{owner}/{repo_name}/contents/{path}"

    # Send a GET request to the API
    response = httpx.get(api_url)

    # Check if the request was successful
    if response.is_error:
        console.print(
            f"[bold red]Failed to list folders.[/bold red] Status Code: {response.status_code}, Error: {response.text}"
        )
        exit(1)

    contents = response.json()
    folders = [item["name"] for item in contents if item["type"] == "dir"]
    return folders


@cli_start.command(name="list")
def list_integrations() -> None:
    """
    List all available public integrations.
    """
    console.print("🌊 Here are the integrations available to you:", style="bold")
    options = list_git_folders("port-labs", "port-ocean", "integrations")

    for option in options:
        console.print(f"⚓️ [bold][blue]{option}[/blue][/bold]")



from .list_integrations import list_git_folders
from .main import cli_start
from .new import new
from .pull import pull
from .sail import sail
from .version import version

__all__ = [
    "cli_start",
    "list_git_folders",
    "new",
    "pull",
    "sail",
    "version",
]



from typing import Any

from port_ocean.context.ocean import ocean


# Required
# Listen to the resync event of all the kinds specified in the mapping inside port.
# Called each time with a different kind that should be returned from the source system.
@ocean.on_resync()
async def on_resync(kind: str) -> list[dict[Any, Any]]:
    # 1. Get all data from the source system
    # 2. Return a list of dictionaries with the raw data of the state to run the core logic of the framework for
    # Example:
    # if kind == "project":
    #     return [{"some_project_key": "someProjectValue", ...}]
    # if kind == "issues":
    #     return [{"some_issue_key": "someIssueValue", ...}]
    return []


# The same sync logic can be registered for one of the kinds that are available in the mapping in port.
# @ocean.on_resync('project')
# async def resync_project(kind: str) -> list[dict[Any, Any]]:
#     # 1. Get all projects from the source system
#     # 2. Return a list of dictionaries with the raw data of the state
#     return [{"some_project_key": "someProjectValue", ...}]
#
# @ocean.on_resync('issues')
# async def resync_issues(kind: str) -> list[dict[Any, Any]]:
#     # 1. Get all issues from the source system
#     # 2. Return a list of dictionaries with the raw data of the state
#     return [{"some_issue_key": "someIssueValue", ...}]


# Optional
# Listen to the start event of the integration. Called once when the integration starts.
@ocean.on_start()
async def on_start() -> None:
    # Something to do when the integration starts
    # For example create a client to query 3rd party services - GitHub, Jira, etc...
    print("Starting integration")
